from django.db.models import OuterRef
from django.db.models import Prefetch
from django.db.models import Subquery

from aio_client.base import const as aio_base_const
from aio_client.provider.api import PostProviderRequest
from aio_client.provider.api import push_request
from kinder.core.children.models import ChildrenDelegate
from kinder.core.declaration.models import Declaration
from kinder.core.direct.models import Direct
from kinder.core.direct.models import DRS

from concentrator.smev3_v321.base import constants as base_enums
from concentrator.smev3_v321.base.exceptions import ContentFailure
from concentrator.smev3_v321.base.utils import BaseExecutor
from concentrator.smev3_v321.base.utils import SMEV3Response
from concentrator.smev3_v321.base.utils import render_type2xml
from concentrator.smev3_v321.base.utils import get_person_identity_doc_info
from concentrator.smev3_v321.base.utils import get_adaptation_program
from concentrator.smev3_v321.base.utils import get_schedule
from concentrator.smev3_v321.base.utils import get_language
from concentrator.smev3_v321.base.utils import get_child_info
from concentrator.smev3_v321.base.utils import (
    get_medical_report_without_files)
from concentrator.smev3_v321.base.utils import get_person_info
from concentrator.smev3_v321.base.utils import get_address
from concentrator.smev3_v321.constants import ERROR_ORG_CODE
from concentrator.smev3_v321.service_types import kinder_conc

from ..helpers import get_declaration_by_client_or_portal_id


class GetApplicationAdmissionRequestExecutor(BaseExecutor):
    """Исполнитель сервиса GetApplicationAdmissionRequest."""

    name_service = 'GetApplicationAdmissionRequest'
    type_service = kinder_conc.GetApplicationAdmissionRequestType

    @classmethod
    def prepare_query(cls):
        """Выполняет подготовку запроса."""

        subquery_first_delegate = (
            ChildrenDelegate.objects.filter(
                id=Subquery(
                    ChildrenDelegate.objects.filter(
                        children_id=OuterRef('children_id')
                    ).order_by('id').values('id')[:1])
            ).select_related('delegate'))

        subquery_unit_code = (
            Subquery(Direct.objects.filter(
                declaration_id=OuterRef('id'),
                status__code__in=(DRS.REGISTER, DRS.DOGOVOR),
                group__unit__is_not_show_on_poral=False
            ).values('group__unit__code')[:1]))

        query = (
            Declaration.objects.annotate(
                unit_code=subquery_unit_code
            ).select_related(
                'children'
            ).filter(
                unit_code__isnull=False
            ).prefetch_related(
                Prefetch(
                    'children__childrendelegate_set',
                    queryset=subquery_first_delegate,
                    to_attr='first_childrendelegate')))

        return query

    @classmethod
    def process(cls, message, request_body):

        declaration, _ = get_declaration_by_client_or_portal_id(
            cls.prepare_query(),
            request_body.GetApplicationAdmissionRequest.orderId)

        try:
            response_body = render_type2xml(
                cls.get_response(declaration), name_type='FormDataResponse')
            content_failure_comment = None
        except ContentFailure as exc:
            response_body = render_type2xml(
                kinder_conc.FormDataResponseType(
                    changeOrderInfo=kinder_conc.changeOrderInfoType(
                        orderId=kinder_conc.orderIdType(
                            request_body.GetApplicationAdmissionRequest.orderId
                        ),
                        statusCode=kinder_conc.statusCodeType(ERROR_ORG_CODE),
                        comment=exc.content_failure_comment)),
                name_type='FormDataResponse')
            content_failure_comment = exc.content_failure_comment

        response = push_request(PostProviderRequest(
            origin_message_id=message['origin_message_id'],
            body=response_body,
            message_type=message['message_type'],
            replay_to=message['replay_to'],
            is_test_message=message['is_test_message']))

        return SMEV3Response(
            response,
            {
                'method_name': cls.name_service,
                'response': response_body,
                'result': content_failure_comment or base_enums.SUCCESS_MESSAGE
            })

    @classmethod
    def get_response(cls, declaration):
        """Формирует ответ на GetApplicationAdmissionRequest.

        :param declaration: Заявление
        :type declaration: Declaration
        :return: ответ
        :rtype: FormDataResponseType
        :raise: ContentFailure
        """

        if not declaration:
            raise ContentFailure(
                aio_base_const.FAILURE,
                'Заявление по указанным параметрам не найдено')

        child = declaration.children
        children_delegate, *_ = child.first_childrendelegate or (None,)
        if not children_delegate:
            raise ContentFailure(
                aio_base_const.FAILURE, 'Не найден представитель')

        delegate = children_delegate.delegate

        entry_params = kinder_conc.EntryParamsType(
            EntryDate=declaration.desired_date,
            Language=kinder_conc.DataElementType(
                **get_language(declaration.spec)),
            Schedule=kinder_conc.DataElementType(
                **get_schedule(declaration.work_type)),
            AgreementOnFullDayGroup=declaration.consent_full_time_group)

        get_application_admission_response = (
            kinder_conc.GetApplicationAdmissionResponseType(
                orderId=int(declaration.client_id),
                PersonInfo=get_person_info(kinder_conc, delegate),
                PersonIdentityDocInfo=(
                    get_person_identity_doc_info(
                        kinder_conc, delegate, declaration)),
                ChildInfo=get_child_info(kinder_conc, declaration),
                Address=get_address(kinder_conc, child),
                EntryParams=entry_params,
                AdaptationProgram=(
                    get_adaptation_program(kinder_conc, declaration)),
                MedicalReport=get_medical_report_without_files(
                    kinder_conc, declaration),
                EduOrganizationCode=declaration.unit_code))

        response = kinder_conc.FormDataResponseType(
            GetApplicationAdmissionResponse=get_application_admission_response)

        return response
